@@ -1,5 +1,5 @@
 /* Support for printing Pascal types for GDB, the GNU debugger.
-   Copyright (C) 2000-2022 Free Software Foundation, Inc.
+   Copyright (C) 2000-2025 Free Software Foundation, Inc.
 
    This file is part of GDB.
 
@@ -18,9 +18,8 @@
 
 /* This file is derived from p-typeprint.c */
 
-#include "defs.h"
-#include "gdbsupport/gdb_obstack.h"
-#include "bfd.h"		/* Binary File Description */
+#include "event-top.h"
+#include "bfd.h"
 #include "symtab.h"
 #include "gdbtypes.h"
 #include "expression.h"
@@ -31,7 +30,6 @@
 #include "p-lang.h"
 #include "typeprint.h"
 #include "gdb-demangle.h"
-#include <ctype.h>
 #include "cli/cli-style.h"
 
 /* See language.h.  */
@@ -55,7 +53,8 @@ pascal_language::print_type (struct type *type, const char *varstring,
       type_print_varspec_prefix (type, stream, show, 0, flags);
     }
   /* first the name */
-  gdb_puts (varstring, stream);
+  if (varstring != nullptr)
+    gdb_puts (varstring, stream);
 
   if ((varstring != NULL && *varstring != '\0')
       && !(code == TYPE_CODE_FUNC
@@ -138,13 +137,13 @@ pascal_language::type_print_method_args (const char *physname,
     {
       gdb_puts (" (", stream);
       /* We must demangle this.  */
-      while (isdigit (physname[0]))
+      while (c_isdigit (physname[0]))
 	{
 	  int len = 0;
 	  int i, j;
 	  char *argname;
 
-	  while (isdigit (physname[len]))
+	  while (c_isdigit (physname[len]))
 	    {
 	      len++;
 	    }
@@ -238,35 +237,12 @@ pascal_language::type_print_varspec_prefix (struct type *type,
 	gdb_printf (stream, "(");
       gdb_printf (stream, "array ");
       if (type->target_type ()->length () > 0
-	  && type->bounds ()->high.kind () != PROP_UNDEFINED)
+	  && type->bounds ()->high.is_constant ())
 	gdb_printf (stream, "[%s..%s] ",
 		    plongest (type->bounds ()->low.const_val ()),
 		    plongest (type->bounds ()->high.const_val ()));
       gdb_printf (stream, "of ");
       break;
-
-    case TYPE_CODE_UNDEF:
-    case TYPE_CODE_STRUCT:
-    case TYPE_CODE_UNION:
-    case TYPE_CODE_ENUM:
-    case TYPE_CODE_INT:
-    case TYPE_CODE_FLT:
-    case TYPE_CODE_VOID:
-    case TYPE_CODE_ERROR:
-    case TYPE_CODE_CHAR:
-    case TYPE_CODE_BOOL:
-    case TYPE_CODE_SET:
-    case TYPE_CODE_RANGE:
-    case TYPE_CODE_STRING:
-    case TYPE_CODE_COMPLEX:
-    case TYPE_CODE_TYPEDEF:
-    case TYPE_CODE_FIXED_POINT:
-      /* These types need no prefix.  They are listed here so that
-	 gcc -Wall will reveal any types that haven't been handled.  */
-      break;
-    default:
-      gdb_assert_not_reached ("unexpected type");
-      break;
     }
 }
 
@@ -377,29 +353,6 @@ pascal_language::type_print_varspec_suffix (struct type *type,
       type_print_func_varspec_suffix (type, stream, show,
 					     passed_a_ptr, 0, flags);
       break;
-
-    case TYPE_CODE_UNDEF:
-    case TYPE_CODE_STRUCT:
-    case TYPE_CODE_UNION:
-    case TYPE_CODE_ENUM:
-    case TYPE_CODE_INT:
-    case TYPE_CODE_FLT:
-    case TYPE_CODE_VOID:
-    case TYPE_CODE_ERROR:
-    case TYPE_CODE_CHAR:
-    case TYPE_CODE_BOOL:
-    case TYPE_CODE_SET:
-    case TYPE_CODE_RANGE:
-    case TYPE_CODE_STRING:
-    case TYPE_CODE_COMPLEX:
-    case TYPE_CODE_TYPEDEF:
-    case TYPE_CODE_FIXED_POINT:
-      /* These types do not need a suffix.  They are listed so that
-	 gcc -Wall will report types that may not have been considered.  */
-      break;
-    default:
-      gdb_assert_not_reached ("unexpected type");
-      break;
     }
 }
 
@@ -532,7 +485,9 @@ pascal_language::type_print_base (struct type *type, struct ui_file *stream, int
 
 	      if (HAVE_CPLUS_STRUCT (type))
 		{
-		  if (TYPE_FIELD_PROTECTED (type, i))
+		  field &fld = type->field (i);
+
+		  if (fld.is_protected ())
 		    {
 		      if (section_type != s_protected)
 			{
@@ -541,7 +496,7 @@ pascal_language::type_print_base (struct type *type, struct ui_file *stream, int
 				      level + 2, "");
 			}
 		    }
-		  else if (TYPE_FIELD_PRIVATE (type, i))
+		  else if (fld.is_private ())
 		    {
 		      if (section_type != s_private)
 			{
@@ -562,21 +517,20 @@ pascal_language::type_print_base (struct type *type, struct ui_file *stream, int
 		}
 
 	      print_spaces (level + 4, stream);
-	      if (field_is_static (&type->field (i)))
+	      if (type->field (i).is_static ())
 		gdb_printf (stream, "static ");
 	      print_type (type->field (i).type (),
 				 type->field (i).name (),
 				 stream, show - 1, level + 4, flags);
-	      if (!field_is_static (&type->field (i))
-		  && TYPE_FIELD_PACKED (type, i))
+	      if (!type->field (i).is_static ()
+		  && type->field (i).is_packed ())
 		{
 		  /* It is a bitfield.  This code does not attempt
 		     to look at the bitpos and reconstruct filler,
 		     unnamed fields.  This would lead to misleading
 		     results if the compiler does not put out fields
 		     for such things (I don't know what it does).  */
-		  gdb_printf (stream, " : %d",
-			      TYPE_FIELD_BITSIZE (type, i));
+		  gdb_printf (stream, " : %d", type->field (i).bitsize ());
 		}
 	      gdb_printf (stream, ";\n");
 	    }
